/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#include "perm_permissions_table.h"
#include "access_t.h"
#include "proc_operations.h"
#include "process_info.h"
#include "set_mode_t.h"
#include <fcntl.h>
#include <pthread.h>
#include <sqlite3.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/fsuid.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

// Global pointer to the SQLite database storing permanent permissions
sqlite3 *perm_database = NULL;
const char *const table_name = "permissions";
// Each row represents a permission to access a specific file for an executable
const int column_count = 3;
// Column names in the permissions table
const char *const schema[] = {"executable", "filename", "mode"};
// Expected SQL data types for each column
const char *const types[] = {"TEXT", "TEXT", "INTEGER"};
uid_t ruid, euid, current_uid;

/**
 * Callback to validate the database schema matches expectations.
 * Verifies column order, names, and types during schema checks.
 *
 * @param unused: Ignored context pointer
 * @param argc: Number of columns returned (should be 3)
 * @param argv: Column data: index 0=name, 1=type, 2=notnull flag
 * @param colname: Column names (unused)
 * @return 0 on success, 1 if schema mismatch
 */
static int check_table_col_schema(void *notused, int argc, char **argv,
                                  char **colname) {
  (void)notused;
  (void)colname;
  if (argc < 3) {
    fprintf(stderr,
            "[ICFS] Unexpected amount of arguments given to the callback.\n");
    return 1;
  }
  int column_num = atoi(argv[0]);
  if (column_num >= column_count) {
    fprintf(stderr, "[ICFS] Table contains unexpected amount of columns.\n");
    return 1;
  }

  if (strcmp(schema[column_num], argv[1]) == 0 &&
      strcmp(types[column_num], argv[2]) == 0) {
    return 0;
  }
  fprintf(stderr, "[ICFS] Column %d does not conform to the schema.\n",
          column_num);
  return 1;
}

/**
 * Creates the permissions table schema in the database.
 * Called when the table doesn't exist yet.
 *
 * @return 0 on success, 1 on failure
 */
int create_database_schema(void) {
  fprintf(stderr, "[ICFS] Creating table 'permissions'.\n");
  const char *create_query =
      "CREATE TABLE permissions(executable TEXT NOT "
      "NULL, filename TEXT NOT NULL, mode INTEGER NOT NULL);";
  char *err = NULL;
  int ret = sqlite3_exec(perm_database, create_query, NULL, NULL, &err);

  if (ret != SQLITE_OK) {
    fprintf(stderr, "[ICFS] sqlite3 error: %s\n", err);
    sqlite3_free(err);
    return 1;
  }

  fprintf(stderr, "[ICFS] Database created successfully\n");
  return 0;
}

/**
 * Ensures the database schema matches expected structure.
 * Validates table existence and column definitions.
 *
 * @return 0 if schema is valid, 1 if validation failed or repair failed
 */
int ensure_database_schema(void) {
  // Check for the table.
  int result = sqlite3_table_column_metadata(
      perm_database, NULL, table_name, NULL, NULL, NULL, NULL, NULL, NULL);
  if (result == SQLITE_ERROR) {
    fprintf(stderr, "[ICFS] Table '%s' does not exist.\n", table_name);
    if (create_database_schema()) {
      fprintf(stderr, "[ICFS] Table could not be created.\n");
      return 1;
    }
    return 0;
  } else if (result != SQLITE_OK) {
    fprintf(stderr, "[ICFS] Database metadata could not be retrieved.\n");
    return 1;
  }

  // Verify column definitions
  const char *pragma = "PRAGMA table_info(permissions);";
  char *err = NULL;
  int ret =
      sqlite3_exec(perm_database, pragma, check_table_col_schema, NULL, &err);

  if (ret != SQLITE_OK) {
    fprintf(stderr, "[ICFS] sqlite3 error: %s\n", err);
    sqlite3_free(err);
    return 1;
  }

  fprintf(stderr, "[ICFS] Schema is correct.\n");
  return 0;
}

/**
 * Initializes the permanent permissions database.
 * Creates/opens the database file and sets up schema.
 *
 * @param db_filename Path to the SQLite database file
 * @return 0 on success, -1 on failure
 */
int init_perm_permissions_table(const char *db_filename) {
  // Prevent group/others access during file creation
  umask(0077);
  ruid = getuid();
  euid = geteuid();
  fprintf(stderr, "[ICFS] Running with uid: %d, gid: %d\n", euid, getegid());

  // Open database with read/write access and full mutex protection
  if (sqlite3_open_v2(db_filename, &perm_database,
                      SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE |
                          SQLITE_OPEN_FULLMUTEX,
                      NULL)) {
    perror("[ICFS] Can't open permanent permissions database");
    return -1;
  }
  umask(0); // Restore default umask

  // Verify and initialize schema
  if (ensure_database_schema()) {
    fprintf(stderr, "[ICFS] Database schema is not correct.\n");
    return -1;
  }

  // Switch to real UID, since we started with the icfs user to open the
  // database
  int status = seteuid(ruid);
  if (status < 0) {
    fprintf(stderr,
            "[ICFS] Couldn't set euid to ruid during database setup.\n");
    exit(status);
  }

  return 0;
}

/**
 * Closes the database connection and releases resources.
 * Should be called when the module is shutting down.
 */
void destroy_perm_permissions_table(void) { sqlite3_close(perm_database); }

/**
 * Checks if a specific process has permanent access to a file.
 * Does not check parent processes.
 *
 * @param filename Path to the file being accessed
 * @param pi Process information structure
 * @return ALLOW (explicit allow), DENY (explicit deny), or NDEF (no info)
 */
access_t check_perm_access_noparent(const char *filename,
                                    struct process_info pi) {
  if (pi.name == NULL)
    return NDEF;

  access_t ret = NDEF;
  sqlite3_stmt *stmt = NULL;

  // Query checks:
  // 1. Exact match for filename
  // 2. Directory prefix match (filename starts with stored path + '/')
  // Ordered by longest path for most specific match first
  const char *sql =
      "SELECT mode FROM permissions WHERE executable = ?1 "
      "AND (( ?2 LIKE (filename || \'%\') AND filename "
      "GLOB \'*/\') OR filename = ?2 ) ORDER BY LENGTH( filename ) DESC;";

  sqlite3_prepare_v2(perm_database, sql, -1, &stmt, NULL);
  sqlite3_bind_text(stmt, 1, pi.name, -1, SQLITE_STATIC);
  sqlite3_bind_text(stmt, 2, filename, -1, SQLITE_STATIC);

  int step_ret = sqlite3_step(stmt);
  if (step_ret != SQLITE_ROW && step_ret != SQLITE_DONE) {
    fprintf(stderr, "[ICFS] SQLite error: %s\n", sqlite3_errstr(step_ret));
    sqlite3_finalize(stmt);
    return ret;
  }

  if (step_ret == SQLITE_ROW) {
    int mode_col = sqlite3_column_int(stmt, 0);
    ret = mode_col ? ALLOW : DENY;
  }

  sqlite3_finalize(stmt);
  return ret;
}

/**
 * Checks if a process or any of its ancestors have permanent access.
 * Handles hierarchical permission inheritance.
 *
 * @param filename Path to the file being accessed
 * @param pi Process information structure
 * @return ALLOW/DENY/NDEF with NDEF meaning no explicit rule found
 * @note May return false negatives if parent processes terminate during check
 */
access_t check_perm_access(const char *filename, struct process_info pi) {
  if (pi.PID == 0 || pi.name == NULL) {
    return NDEF;
  }

  struct process_info current_pi = pi;
  current_pi.name = strdup(current_pi.name);

  while (current_pi.PID != 0) {
    access_t access = check_perm_access_noparent(filename, current_pi);
    free(current_pi.name);

    if (access != NDEF) {
      return access;
    }

    current_pi.name = NULL;
    // Traverse to parent process
    while (current_pi.name == NULL) {
      current_pi.PID = get_main_thread_pid(get_parent_pid(current_pi.PID));
      if (current_pi.PID != 0) {
        current_pi.name = get_process_name_by_pid(current_pi.PID);
      } else {
        break;
      }
    }
  }

  return NDEF;
}

/**
 * Sets a permanent access rule for a process-file combination.
 *
 * @param filename Path to the file needing access
 * @param pi Process information
 * @param mode Kind of access rule to be set - SET_DENY to deny access, and
 * SET_ALLOW to allow access.
 * @return 0 on success, 1 on failure
 */
int set_perm_access(const char *filename, struct process_info pi,
                    set_mode_t mode) {
  sqlite3_stmt *stmt = NULL;
  char *sql = NULL;

  if (mode == SET_ALLOW) {
    sql = "INSERT INTO permissions VALUES (?1, ?2, TRUE);";
  } else if (mode == SET_DENY) {
    sql = "INSERT INTO permissions VALUES (?1, ?2, FALSE);";
  } else {
    return 1; // Invalid mode
  }

  sqlite3_prepare_v2(perm_database, sql, -1, &stmt, NULL);
  sqlite3_bind_text(stmt, 1, pi.name, -1, SQLITE_STATIC);
  sqlite3_bind_text(stmt, 2, filename, -1, SQLITE_STATIC);

  int step_ret = sqlite3_step(stmt);
  if (step_ret != SQLITE_DONE) {
    fprintf(stderr, "[ICFS] SQLite error: %s\n", sqlite3_errstr(step_ret));
    sqlite3_finalize(stmt);
    return 1;
  }

  sqlite3_finalize(stmt);
  return 0;
}
