/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#ifndef PERM_PERMISSION_TABLE_H
#define PERM_PERMISSION_TABLE_H

#include "access_t.h"
#include "process_info.h"
#include "set_mode_t.h"

/**
 * Initializes the permanent permissions database.
 * Creates/opens the database file and sets up schema.
 *
 * @param db_filename Path to the SQLite database file
 * @return 0 on success, -1 on failure
 */
int init_perm_permissions_table(const char *db_filename);

/**
 * Closes the database connection and releases resources.
 * Should be called when the module is shutting down.
 */
void destroy_perm_permissions_table();

/**
 * Checks if a process or any of its ancestors have permanent access.
 * Handles hierarchical permission inheritance.
 *
 * @param filename Path to the file being accessed
 * @param pi Process information structure
 * @return ALLOW/DENY/NDEF with NDEF meaning no explicit rule found
 * @note May return false negatives if parent processes terminate during check
 */
access_t check_perm_access(const char *filename, struct process_info pi);

/**
 * Sets a permanent access rule for a process-file combination.
 *
 * @param filename Path to the file needing access
 * @param pi Process information
 * @param mode SET_ALLOW (whitelist) or SET_DENY (blacklist)
 * @return 0 on success, 1 on failure
 */
int set_perm_access(const char *filename, struct process_info pi,
                    set_mode_t mode);

#endif // #ifdef PERM_PERMISSION_TABLE_H
