/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#ifndef PROC_OPERATIONS
#define PROC_OPERATIONS

#include <time.h>

/**
 * @brief Retrieves the full path of the executable for a given process ID.
 *
 * This function reads the `/proc/<pid>/exe` symbolic link, which points to the
 * executable file of the process. The returned string is dynamically allocated
 * and must be freed by the caller.
 *
 * @param pid The process ID to query.
 * @return char* Dynamically allocated string containing the executable path, or
 * NULL on failure.
 */
char *get_process_name_by_pid(const int pid);

/**
 * @brief Finds the parent process ID of a given process.
 *
 * This function parses the `/proc/<pid>/status` file to extract the `PPid`
 * field, which represents the parent process ID. This is useful for tracing
 * process lineage.
 *
 * @param pid The process ID of the process to find the parent of.
 * @return pid_t The parent process ID, or 0 if the parent could not be
 * determined.
 */
pid_t get_parent_pid(pid_t pid);

/**
 * @brief Returns the PID of the main thread (i.e., the process ID) of the
 * process that the given thread ID (tid) belongs to.
 *
 * In Linux, threads within the same process share the same thread group ID
 * (TGID), which is equal to the PID of the main thread. This function retrieves
 * the TGID by parsing `/proc/<tid>/status`, effectively returning the main
 * thread's PID.
 *
 * @param tid The thread ID (TID) of any thread in the process.
 * @return pid_t The process ID (main thread's PID), or 0 on error (invalid tid
 * or file read failure).
 */
pid_t get_main_thread_pid(pid_t tid);

#endif // !PROC_OPERATIONS
