/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#ifndef SOURCEFS_H
#define SOURCEFS_H

#include <dirent.h>
#include <sys/stat.h>

/**
 * @brief Initialize the source filesystem with a root directory
 *
 * Sets up the mount point and opens a protected file descriptor to the
 * root directory for safe relative operations.
 *
 * @param root_path Absolute path to the physical root directory
 * @return 0 on success, -1 on failure
 */
int source_init(const char *root_path);

/**
 * @brief Clean up resources used by the source filesystem
 */
void source_destroy(void);

/**
 * @brief Get file status information
 *
 * Uses fstatat() to retrieve metadata relative to root_fd. Follows symlinks
 * by default (flags=0).
 */
int source_stat(const char *restrict filename, struct stat *restrict statbuf);

/**
 * @brief Open a directory for reading
 *
 * Combines openat() and fdopendir() to safely access directories relative
 * to the root_fd.
 */
DIR *source_opendir(const char *filename);

/**
 * @brief Remove a file from the filesystem
 *
 * Uses unlinkat() with AT_REMOVEDIR flag to safely remove files relative
 * to the root directory.
 */
int source_unlink(const char *filename);

/**
 * @brief Create a directory in the filesystem
 *
 * Uses mkdirat() to safely create directories relative to root_fd,
 * preventing race conditions from concurrent path modifications.
 */
int source_mkdir(const char *filename, mode_t mode);

/**
 * @brief Remove an empty directory
 *
 * Uses unlinkat() with AT_REMOVEDIR flag to safely remove directories.
 */
int source_rmdir(const char *filename);

/**
 * @brief Create a symbolic link
 *
 * Creates symlinks relative to the root_fd directory for safety.
 */
int source_symlink(const char *target, const char *linkpath);

/**
 * @brief Rename a file or directory
 *
 * Uses renameat() to safely rename within the same root_fd namespace.
 */
int source_rename(const char *oldpath, const char *newpath);

/**
 * @brief Create a hard link
 *
 * Uses linkat() with flags=0 (default behavior). May need AT_SYMLINK_NOFOLLOW
 * if symlink handling should be modified.
 */
int source_link(const char *oldpath, const char *newpath);

/**
 * @brief Change file access mode
 *
 * Uses fchmodat() with flags=0 (follow symlinks). Consider using
 * AT_SYMLINK_NOFOLLOW if symlink metadata should be modified directly.
 */
int source_chmod(const char *filename, mode_t mode);

/**
 * @brief Change file owner and group
 *
 * Uses fchownat() with AT_SYMLINK_NOFOLLOW to modify symlink metadata
 * rather than its target.
 */
int source_chown(const char *filename, uid_t owner, gid_t group);

/**
 * @brief Truncate a file to a specified length
 *
 * Opens the file with read-only access then truncates it. This may fail
 * if the file wasn't opened with write permissions. Consider changing
 * openat() flags to O_WRONLY for reliability.
 */
int source_truncate(const char *filename, off_t length);

/**
 * @brief Check file access permissions
 *
 * Uses faccessat() to check access rights relative to root_fd.
 */
int source_access(const char *filename, int mode);

/* `open` and `create` are designed to correspond to fuse operations, not the
 * libc's `open(2)`. Both of them actually call `openat`. */

/**
 * @brief Open a file with specified flags
 *
 * Uses openat() to safely access files relative to root_fd.
 */
int source_open(const char *filename, int flags);

/**
 * @brief Create and open a new file
 *
 * Uses openat() with O_CREAT to create files relative to root_fd.
 */
int source_create(const char *filename, int flags, mode_t mode);

#endif // !SOURCEFS_H
