/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#include "temp_permissions_table.h"
#include "access_t.h"
#include "cc.h"
#include "proc_operations.h"
#include "process_info.h"
#include <pthread.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#define GC_INTERVAL 1 // Garbage collection time interval.

/**
 * @struct temp_process_permissions
 * @brief Stores temporary file access permissions for a process
 *
 * This structure tracks temporary file access permissions for a specific
 * process. The permissions are tied to the process's lifetime using its
 * creation time.
 */
struct temp_process_permissions {
  unsigned long long creation_time; /**< Process creation time */
  vec(char *) allowed_files; /**< List of allowed file paths (prefixes) */
  vec(char *) denied_files;  /**< List of denied file paths (prefixes) */
};

/**
 * @brief Global table mapping PIDs to temporary permissions
 *
 * This map stores the temporary file access permissions for processes.
 * Entries are keyed by process ID (pid_t) and contain temp_process_permissions
 * structures with the access rules.
 */
map(pid_t, struct temp_process_permissions) temp_permissions_table;

/**
 * @brief Read-write lock for thread-safe access to temp_permissions_table
 *
 * A read-write lock to ensure thread-safe operations on the
 * temp_permissions_table. Allows concurrent reads but exclusive writes.
 */
pthread_rwlock_t temp_permissions_table_lock = PTHREAD_RWLOCK_INITIALIZER;

/**
 * @brief Thread handle for the garbage collector
 *
 * This thread handles cleanup of stale entries in the temp_permissions_table.
 */
pthread_t gc_thread;

/**
 * @brief Flag indicating whether garbage collector is active
 *
 * When non-zero, indicates that the garbage collector thread should continue
 * running.
 */
int is_gc_active = 0;

/**
 * Function to get the process creation time from the proc
 * filesystem
 *
 * @param pid The process ID of the process to get the creation time of
 * @return The process creation time, or 0 on error
 * @note although nothing in the documentation says that the creation time is
 * never really equal to 0, it exceptionally unlikely.
 */
unsigned long long get_process_creation_time(pid_t pid) {
  char path[256];
  FILE *fp = NULL;
  unsigned long long creation_time = 0;

  // Construct the path to the process's status file
  snprintf(path, sizeof(path), "/proc/%u/stat", pid);

  // Open the status file
  fp = fopen(path, "r");
  if (fp == NULL) {
    perror("[ICFS] fopen");
    return 0;
  }

  // Read the creation time (the 22nd field in the stat file)
  for (int i = 1; i < 22; i++) {
    if (fscanf(fp, "%*s") == EOF) {
      fprintf(stderr,
              "[ICFS] Error reading process stat file on the number %d\n", i);
      fclose(fp);
      return 0;
    }
  }
  if (fscanf(fp, "%llu", &creation_time) != 1) {
    fprintf(stderr, "[ICFS] Error reading creation time\n");
    fclose(fp);
    return 0;
  }

  // Close the file
  fclose(fp);

  return creation_time;
}

/**
 * @brief Validates if a process entry is still valid
 *
 * Checks if the given process entry matches the current process state
 * by comparing creation times.
 *
 * @param pid Process ID to validate
 * @param entry Pointer to the permission entry to validate
 * @return 1 if valid, 0 if invalid or error
 */
int is_valid(pid_t pid, struct temp_process_permissions *entry) {
  unsigned long long creation_time = get_process_creation_time(pid);
  if (creation_time == 0) {
    return 0;
  }

  // If the creation time doesn't match, this is a different process instance
  if (creation_time != entry->creation_time) {
    return 0;
  }

  return 1;
}

/**
 * @brief Garbage collector thread for cleaning up stale entries
 *
 * Periodically scans the permissions table and removes entries for processes
 * that no longer exist. Runs every second while is_gc_active is true.
 *
 * @param arg Unused thread argument parameter
 * @return Always NULL
 */
void *garbage_collector(void *arg) {
  (void)arg;

  while (is_gc_active) {
    sleep(GC_INTERVAL); // Check once per second for stale entries
    pthread_rwlock_wrlock(&temp_permissions_table_lock);

    vec(pid_t) blacklist;
    init(&blacklist);

    // Identify stale entries
    for_each(&temp_permissions_table, pid, entry) {
      if (!is_valid(*pid, entry)) {
        push(&blacklist, *pid);
        // Free memory for this entry's file lists
        for_each(&entry->allowed_files, allowed_file) { free(*allowed_file); }
        cleanup(&entry->allowed_files);
        for_each(&entry->denied_files, denied_file) { free(*denied_file); }
        cleanup(&entry->denied_files);
      }
    }

    // Remove all stale entries from the table
    for_each(&blacklist, pid) { erase(&temp_permissions_table, *pid); }

    cleanup(&blacklist);

    pthread_rwlock_unlock(&temp_permissions_table_lock);
  }

  return NULL;
}

/**
 * Initializes the temporary permissions table.
 *
 * @return 0 on success, -1 on failure
 */
int init_temp_permissions_table(void) {
  init(&temp_permissions_table);
  return 0;
}
/**
 * Starts the temporary permissions table garbage_collector.
 *
 * @return 0 on success, -1 on failure
 */
int init_garbage_collector(void) {
  is_gc_active = 1;
  if (pthread_create(&gc_thread, NULL, garbage_collector, NULL) != 0) {
    return -1;
  }
  return 0;
}

/**
 * Destroys the temporary permissions table.
 *
 * @note the table is guranteed to be destroyed if it is already initialized.
 * It does not indicate any errors whatsoever. If something goes wrong - you are
 * screwed.
 */
void destroy_temp_permissions_table(void) {
  if (is_gc_active) {
    is_gc_active = 0;
    pthread_join(gc_thread, NULL);
  }

  // Free the memory allocated for the table
  for_each(&temp_permissions_table, entry) {
    for_each(&entry->allowed_files, allowed_file) { free(*allowed_file); }
    cleanup(&entry->allowed_files);
  }
  for_each(&temp_permissions_table, entry) {
    for_each(&entry->denied_files, denied_file) { free(*denied_file); }
    cleanup(&entry->denied_files);
  }

  cleanup(&temp_permissions_table);
}

/**
 * @brief Checks if a specific process has temporary access to a file
 *
 * Checks only the specified process (not its parents) for temporary access
 * permissions to the given file. Uses a longest-match algorithm for path
 * prefixes.
 *
 * @param filename Path to the file being accessed
 * @param pid Process ID of the process to check
 * @return Access status ALLOW, DENY, or NDEF (no information)
 */
access_t check_temp_access_noparent(const char *filename, pid_t pid) {
  pthread_rwlock_rdlock(&temp_permissions_table_lock);
  struct temp_process_permissions *permission_entry =
      get(&temp_permissions_table, pid);
  if (permission_entry != NULL) {
    unsigned long long process_creation_time = get_process_creation_time(pid);
    if (process_creation_time == 0) {
      perror("[ICFS] Could not retrieve process creation time");
      pthread_rwlock_unlock(&temp_permissions_table_lock);
      return NDEF;
    }

    if (process_creation_time == permission_entry->creation_time) {
      // The process is the same as the one that was granted temporary access
      size_t filename_len = strlen(filename);
      access_t ret = NDEF;
      size_t maxlen = 0;

      // Check denied files first (deny takes precedence over allow)
      for_each(&permission_entry->denied_files, denied_file) {
        size_t denied_file_len = strlen(*denied_file);
        // Check if this denied path prefix matches the requested file
        if ((strncmp(*denied_file, filename, denied_file_len) == 0 &&
             ((denied_file_len < filename_len &&
               (*denied_file)[denied_file_len - 1] == '/') ||
              (denied_file_len == filename_len))) &&
            denied_file_len > maxlen) {
          maxlen = denied_file_len;
          ret = DENY;
        }
      }

      // Check allowed files
      for_each(&permission_entry->allowed_files, allowed_file) {
        size_t allowed_file_len = strlen(*allowed_file);
        // Check if this allowed path prefix matches the requested file
        if ((strncmp(*allowed_file, filename, allowed_file_len) == 0 &&
             ((allowed_file_len < filename_len &&
               (*allowed_file)[allowed_file_len - 1] == '/') ||
              (allowed_file_len == filename_len))) &&
            allowed_file_len > maxlen) {
          maxlen = allowed_file_len;
          ret = ALLOW;
        }
      }

      pthread_rwlock_unlock(&temp_permissions_table_lock);
      return ret;
    }
  }
  pthread_rwlock_unlock(&temp_permissions_table_lock);
  return NDEF;
}

/**
 * Checks if the process or any of it's parents have temporary access to the
 * file.
 *
 * @param filename The file that the process is trying to access
 * @param pi The process information
 * @return access status - ALLOW, DENY or NDEF in case if no information was
 * found. Does not return ALLOW_TEMP.
 * @note In case one of the parent processes is killed while this function
 * execution the result is not guranteed to be correct. It should only
 * false negatives, though.
 */
access_t check_temp_access(const char *filename, struct process_info pi) {
  pid_t current_pid = pi.PID;
  while (current_pid != 0) {
    access_t access = check_temp_access_noparent(filename, current_pid);
    if (access != NDEF) {
      return access;
    }
    current_pid = get_main_thread_pid(get_parent_pid(current_pid));
  }

  return NDEF;
}

/**
 * Sets temporary access mode of the process to the file.
 *
 * @param filename The file that the process is trying to access
 * @param pi The process information
 * @param mode Kind of access rule to be set - SET_DENY to deny access, and
 * SET_ALLOW to allow access.
 * @return 0 on success, -1 on failure.
 */
int set_temp_access(const char *filename, struct process_info pi,
                    set_mode_t mode) {
  if (pi.PID == 0)
    return NDEF;
  pthread_rwlock_wrlock(&temp_permissions_table_lock);
  struct temp_process_permissions *permission_entry =
      get(&temp_permissions_table, pi.PID);

  if (permission_entry != NULL) {

    unsigned long long process_creation_time =
        get_process_creation_time(pi.PID);
    if (process_creation_time == 0) {
      perror("[ICFS] Could not retrieve process creation time");
      pthread_rwlock_unlock(&temp_permissions_table_lock);
      return -1;
    }

    if (process_creation_time == permission_entry->creation_time) {
      // The process is the same as the one that was granted temporary access
      if (mode == SET_ALLOW) {
        push(&permission_entry->allowed_files, strdup(filename));
      }
      if (mode == SET_DENY) {
        push(&permission_entry->denied_files, strdup(filename));
      }

      pthread_rwlock_unlock(&temp_permissions_table_lock);
      return 0;
    }
    // We have an entry for the process, but the process is different
    // Delete the entry and create a new one
    erase(&temp_permissions_table, pi.PID);
    permission_entry = NULL;
  }

  // No entry is present - construct a new one
  struct temp_process_permissions new_permission_entry;

  new_permission_entry.creation_time = get_process_creation_time(pi.PID);
  init(&new_permission_entry.allowed_files);
  init(&new_permission_entry.denied_files);

  if (mode == SET_ALLOW) {
    push(&new_permission_entry.allowed_files, strdup(filename));
  }
  if (mode == SET_DENY) {
    push(&new_permission_entry.denied_files, strdup(filename));
  }

  insert(&temp_permissions_table, pi.PID, new_permission_entry);

  pthread_rwlock_unlock(&temp_permissions_table_lock);
  return 0;
}
