/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#include "access_t.h"
#include <stddef.h>
#include <sys/types.h>
#include <time.h>
#define _GNU_SOURCE // Required for certain POSIX extensions
#include "cc.h"
#include "perm_permissions_table.h"
#include "real_filename.h"
#include "sourcefs.h"
#include "temp_permissions_table.h"
#include "ui-socket.h"
#include <assert.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

// Exit status codes for icfs_dialogue process interaction
#define DIALOGUE_YES 1
#define DIALOGUE_NO 0
#define DIALOGUE_PERM 2
#define DIALOGUE_TEMP 0 // Bitmask position, not value

// Mutex to protect concurrent access to permission tables
pthread_mutex_t access_check_mutex = PTHREAD_MUTEX_INITIALIZER;

// Structure to hold user permission decision response
struct dialogue_response {
  access_t decision;
  char *filename;
};

/**
 * Initialize UI socket and required permission tables
 *
 * @param perm_permissions_db_filename Path to persistent permissions DB
 * @return 0 on success, 1 on failure
 */
int init_ui_socket(const char *perm_permissions_db_filename) {
  FILE *fp = NULL;

  // Initialize in-memory temporary permissions table
  if (init_temp_permissions_table()) {
    fprintf(stderr,
            "[ICFS] Could not initialize temporary permissions table.\n");
    return 1;
  }

  // Initialize permanent permissions database
  if (init_perm_permissions_table(perm_permissions_db_filename)) {
    fprintf(stderr,
            "[ICFS] Could not initialize permanent permissions table.\n");
    return 1;
  }

  // Verify dialogue utility is available
  fp = popen("icfs_dialogue --version", "r");
  if (fp == NULL) {
    perror("[ICFS] Pipe returned an error");
    return 1;
  }

  pclose(fp);
  return 0;
}

/**
 * Clean up UI socket resources
 */
void destroy_ui_socket(void) {
  destroy_temp_permissions_table();
  destroy_perm_permissions_table();
}

/**
 * Query user for file access permission through GUI dialogue
 *
 * Constructs and executes the icfs_dialogue command to get user consent.
 * Handles memory allocation errors gracefully and parses the response.
 *
 * @param filename File being accessed
 * @param proc_info Process requesting access
 * @return Struct containing access decision and resolved filename
 */
struct dialogue_response ask_access(const char *filename,
                                    struct process_info proc_info) {
  FILE *fp = NULL;
  char *command = NULL;
  int ret = asprintf(&command, "icfs_dialogue \"%d\" \"%s\" \"%s\" \"%s\"",
                     proc_info.PID, proc_info.name, get_mountpoint(), filename);

  struct dialogue_response response;
  response.decision = DENY;
  response.filename = NULL;

  if (ret < 0) {
    // Memory allocation failed - create minimal fallback response
    fprintf(stderr, "[ICFS] Could not create query on rule insertion");
    perror("");
    response.filename = malloc(2);
    response.filename[0] = '/';
    response.filename[1] = 0;
    return response;
  }

  // Execute permission dialogue
  fp = popen(command, "r");
  free(command);

  if (fp == NULL) {
    perror("[ICFS] Pipe returned a error");
    response.filename = malloc(2);
    response.filename[0] = '/';
    response.filename[1] = 0;
    return response;
  }

  str(char) dialogue_output;
  init(&dialogue_output);

  char line[1024];

  // Read entire command output
  while (fgets(line, sizeof(line), fp)) {
    push_fmt(&dialogue_output, line);
  }

  int dialogue_exit_code = WEXITSTATUS(pclose(fp));

  fprintf(stderr, "[ICFS] dialogue wrote out %s\n", first(&dialogue_output));
  fprintf(stderr, "[ICFS] dialogue returned %d\n", dialogue_exit_code);

  // Handle empty output case
  if (size(&dialogue_output) == 0) {
    push(&dialogue_output, '/');
  }

  // Allocate and copy final filename
  response.filename = malloc(size(&dialogue_output) + 1);
  strcpy(response.filename, first(&dialogue_output));
  cleanup(&dialogue_output);

  // Parse exit code combination
  if (dialogue_exit_code == (DIALOGUE_YES | DIALOGUE_PERM)) {
    response.decision = ALLOW;
  } else if (dialogue_exit_code == DIALOGUE_YES | DIALOGUE_TEMP) {
    response.decision = ALLOW_TEMP;
  } else if (dialogue_exit_code == (DIALOGUE_NO | DIALOGUE_PERM)) {
    response.decision = DENY;
  } else {
    response.decision = DENY_TEMP;
  }

  return response;
}

/**
 * Determine file access based on permission tables and user input
 *
 * Checks permissions in order:
 * 1. Temporary permission table
 * 2. Permanent permission table
 * 3. User decision via GUI
 *
 * @param filename File being accessed
 * @param pi Process information
 * @param opts Flags to force permission (GRANT_TEMP/GRANT_PERM)
 * @return 1 if access allowed, 0 if denied
 */
int interactive_access(const char *filename, struct process_info proc_info,
                       int opts) {
  char *real_path = real_filename(filename);
  pthread_mutex_lock(&access_check_mutex);

  // First check temporary permissions
  access_t access = check_temp_access(real_path, proc_info);
  if (access == ALLOW) {
    fprintf(stderr, "[ICFS] Permission allowed to %s based on temp table.\n",
            proc_info.name);
    free(real_path);
    pthread_mutex_unlock(&access_check_mutex);
    return 1;
  }
  if (access == DENY) {
    fprintf(stderr, "[ICFS] Permission denied to %s based on temp table.\n",
            proc_info.name);
    free(real_path);
    pthread_mutex_unlock(&access_check_mutex);
    return 0;
  }

  // Then check permanent permissions
  access = check_perm_access(real_path, proc_info);
  if (access == ALLOW) {
    fprintf(stderr, "[ICFS] Permission allowed to %s based on perm table.\n",
            proc_info.name);
    free(real_path);
    pthread_mutex_unlock(&access_check_mutex);
    return 1;
  }
  if (access == DENY) {
    fprintf(stderr, "[ICFS] Permission denied to %s based on perm table.\n",
            proc_info.name);
    free(real_path);
    pthread_mutex_unlock(&access_check_mutex);
    return 0;
  }

  // Handle forced permission grants
  if (opts & GRANT_PERM) {
    fprintf(stderr, "[ICFS] Permission granted permanently to %s.\n",
            proc_info.name);
    set_perm_access(real_path, proc_info, SET_ALLOW);
    free(real_path);
    pthread_mutex_unlock(&access_check_mutex);
    return 1;
  }
  if (opts & GRANT_TEMP) {
    fprintf(stderr, "[ICFS] Permission granted temporarily to %s.\n",
            proc_info.name);
    set_temp_access(real_path, proc_info, SET_ALLOW);
    free(real_path);
    pthread_mutex_unlock(&access_check_mutex);
    return 1;
  }

  // Get user decision
  struct dialogue_response response = ask_access(filename, proc_info);

  // Validate returned filename meets requirements:
  // 1. Must exist
  // 2. Must be prefix of original filename with trailing slash
  //    or exact match
  while (
      source_access(response.filename, F_OK) ||
      !(strncmp(response.filename, filename, strlen(response.filename)) == 0 &&
        ((strlen(response.filename) < strlen(filename) &&
          response.filename[strlen(response.filename) - 1] == '/') ||
         (strlen(response.filename) == strlen(filename))))) {
    fprintf(stderr, "[ICFS] Invalid filename returned by dialogue: %s\n",
            response.filename);
    free(response.filename);
    response = ask_access(filename, proc_info);
  }

  free(real_path);
  real_path = real_filename(response.filename);
  free(response.filename);

  int ret = 0;

  // Apply user decision to appropriate permission table
  if (response.decision == ALLOW) {
    fprintf(stderr,
            "[ICFS] Permission granted permanently to %s based on response.\n",
            proc_info.name);
    set_perm_access(real_path, proc_info, SET_ALLOW);
    ret = 1;
  } else if (response.decision == ALLOW_TEMP) {
    fprintf(stderr,
            "[ICFS] Permission granted temporarily to %s based on response.\n",
            proc_info.name);
    set_temp_access(real_path, proc_info, SET_ALLOW);
    ret = 1;
  } else if (response.decision == DENY_TEMP) {
    fprintf(stderr,
            "[ICFS] Permission denied temporarily to %s based on response.\n",
            proc_info.name);
    set_temp_access(real_path, proc_info, SET_DENY);
    ret = 0;
  } else if (response.decision == DENY) {
    fprintf(stderr,
            "[ICFS] Permission denied permanently to %s based on response.\n",
            proc_info.name);
    set_perm_access(real_path, proc_info, SET_DENY);
    ret = 0;
  }

  pthread_mutex_unlock(&access_check_mutex);
  free(real_path);
  return ret;
}
